""" Changelist Data Storage XML Changelists Data File.
"""
from typing import Generator

from changelist_data.changelist import Changelist
from changelist_data.xml.changelists import changelists_reader
from changelist_data.xml.changelists.changelists_tree import ChangelistsTree


EMPTY_CHANGELISTS_DATA = """<?xml version="1.0" encoding="UTF-8"?>
<changelists>
</changelists>"""


def read_xml(
    changelists_xml: str
) -> list[Changelist]:
    """ Parse the ChangeLists XML file and obtain all ChangeList Data in a list.

**Parameters:**
 - changelists_xml (str): The contents of the ChangeLists file, in xml format.
    
**Returns:**
 list[Changelist] - The list of Changelist objects in the ChangeLists file.
    """
    return list(generate_changelists_from_xml(changelists_xml))


def generate_changelists_from_xml(
    changelists_xml: str
) -> Generator[Changelist, None, None]:
    """ Generates Changelist Data from the Given XML String.
 - Parses the XML string using ElementTree.
 - Creates Changelist data objects one at a time from the ElementTree.

**Parameters:**
 - changelists_xml (str): The Changelists data.xml file contents as a string.

**Yields:**
 Changelist - The Changelist data objects extracted from the XML ElementTree.
    """
    if (cl_manager := changelists_reader.find_changelists_root(
        changelists_reader.parse_xml(changelists_xml)
    )) is None:
        exit("Changelists tag was not found in the xml file.")
    yield from changelists_reader.generate_changelists(cl_manager)


def load_xml(
    changelists_xml: str
) -> ChangelistsTree:
    """ Parse the Changelists XML file into an XML Tree, and Wrap it.

**Returns:**
 ChangelistsTree - An XML Tree changelists interface.
    """
    return ChangelistsTree(
        changelists_reader.parse_xml(changelists_xml)
    )


def new_tree() -> ChangelistsTree:
    """ Create a new Changelists XML Tree, and Wrap it.

**Returns:**
 ChangelistsTree - An XML Tree Changelists interface.
    """
    return ChangelistsTree(
        changelists_reader.parse_xml(EMPTY_CHANGELISTS_DATA)
    )


""" XML Management.
    There are two XML file formats: Changelists and Workspace.
    Each format is handled by it's own package.
    Both Packages provide the same high level methods: read_xml and load_tree
"""



""" Data container for File Change information.
"""
from dataclasses import dataclass


@dataclass(frozen=True)
class FileChange:
    """ The Change Information on a single file.

**Properties:**
 - before_path (str?): The initial path of the file.
 - before_dir (bool?): Whether the initial file is a directory.
 - after_path (str?): The final path of the file.
 - after_dir (bool?): Whether the final path is a directory.
    """
    before_path: str | None = None
    before_dir: bool | None = None
    after_path: str | None = None
    after_dir: bool | None = None


def create_fc(file_path_str: str) -> FileChange:
    """ Build a FC for a created file.
 - Created File Path is stored in the after_path attribute.
 - The after_dir attribute is set to false.

**Parameters:**
 - file_path_str (str): To be stored in FileChange.

**Returns:**
 FileChange - The dataclass object containing a Created FileChange.
    """
    return FileChange(after_path=file_path_str, after_dir=False)


def update_fc(file_path_str: str) -> FileChange:
    """ Build a FC for an updated file.
 - Updated Files contain a before and after path. If they don't match, it's a move-update.
 - Both before_dir and after_dir attribute are set to false.

**Parameters:**
 - file_path_str (str): To be stored in FileChange.

**Returns:**
 FileChange - The dataclass object containing an Update FileChange.
    """
    return FileChange(before_path=file_path_str, before_dir=False, after_path=file_path_str, after_dir=False)


def delete_fc(file_path_str: str) -> FileChange:
    """ Build a FC for a deleted file.
 - Deleted File Path is stored in the before_path attribute.
 - The before_dir attribute is set to false.

**Parameters:**
 - file_path_str (str): To be stored in FileChange.

**Returns:**
 FileChange - The dataclass object containing a Delete FileChange.
    """
    return FileChange(before_path=file_path_str, before_dir=False)


""" Storage for Changelists Data XML Format.
"""
from pathlib import Path
from typing import Generator

from changelist_data.changelist import Changelist
from changelist_data.storage import storage_type, file_validation
from changelist_data.storage.storage_type import StorageType
from changelist_data.xml import changelists
from changelist_data.xml.changelists import ChangelistsTree, EMPTY_CHANGELISTS_DATA


def read_file(
    file_path: Path | None = storage_type.get_default_path(StorageType.CHANGELISTS),
) -> list[Changelist]:
    """ Read a Changelists XML Storage File.
 - Default file_path is given by StorageType.
 - None file_path returns an empty list.

**Parameters:**
 - file_path (Path): The Path to the File containing Changelists XML.

**Returns:**
 list[Changelist] - The list of Changelist data stored in Changelists Storage.
    """
    return list(generate_changelists_from_file(file_path))


def generate_changelists_from_file(
    file_path: Path | None = storage_type.get_default_path(StorageType.CHANGELISTS),
) -> Generator[Changelist, None, None]:
    """ Generate Changelist Data objects from the given Changelist-type Storage file.

**Parameters:**
 - file_path (Path?): The path to the Changelist Data file. Uses default path if None. Default: None.

**Yields:**
 Changelist - The Changelist data objects produced from the storage file.
    """
    if file_path is None:
        return # None Path Given
    if len(file_content := file_validation.validate_file_input_text(file_path)) == 0:
        return # Empty Data File
    yield from changelists.generate_changelists_from_xml(file_content)


def load_file(
    file_path: Path | None = storage_type.get_default_path(StorageType.CHANGELISTS),
) -> ChangelistsTree:
    """ Load a Tree from Changelists XML Storage File.
 - Default file_path

**Parameters:**
 - file_path (Path): The Path to the File containing Changelists XML.

**Returns:**
 ChangelistsTree - The list of Changelist data stored in Changelists Storage.
    """
    if file_path is None or\
        not file_validation.file_exists(file_path):
        return changelists.new_tree()
    # Validate File Stats, Read and Parse the XML
    if len(file_content := file_validation.validate_file_input_text(file_path)) == 0:
        return changelists.new_tree() # Empty Data File
    # Parse and wrap XML ElementTree in ChangelistsTree
    return changelists.load_xml(file_content)


def write_file(
    tree: ChangelistsTree | None,
    file_path: Path = storage_type.get_default_path(StorageType.CHANGELISTS),
) -> bool:
    """ Write a Changelist Data Storage object to an XML File.

**Parameters:**
 - tree (ChangelistDataStorage): The Tree object containing the Changelists.
 - file_path (Path): The File Path to write the XML data to.

**Returns:**
 bool - True after the operation succeeds.
    """
    if tree is None:
        file_path.write_text(EMPTY_CHANGELISTS_DATA)
        return True
    tree.write_tree(file_path)
    return True


""" Storage for Workspace XML Format.
"""
from pathlib import Path
from typing import Generator

from changelist_data.changelist import Changelist
from changelist_data.storage import file_validation, storage_type
from changelist_data.storage.storage_type import StorageType
from changelist_data.xml import workspace
from changelist_data.xml.workspace import read_xml, load_xml
from changelist_data.xml.workspace.workspace_tree import WorkspaceTree


def read_file(
    file_path: Path = storage_type.get_default_path(StorageType.WORKSPACE),
) -> list[Changelist]:
    """ Read a Workspace XML Storage File.
 - Default file_path is given by StorageType.

**Parameters:**
 - file_path (Path): The Path to the File containing Workspace XML.

**Returns:**
 list[Changelist] - The list of Changelist data stored in Workspace Storage.
    """
    return read_xml(
        file_validation.validate_file_input_text(file_path)
    )


def generate_changelists_from_file(
    file_path: Path = storage_type.get_default_path(StorageType.WORKSPACE),
) -> Generator[Changelist, None, None]:
    if len(file_content := file_validation.validate_file_input_text(file_path)) == 0:
        return # Empty Data File
    yield from workspace.generate_changelists_from_xml(file_content)


def load_file(
    file_path: Path = storage_type.get_default_path(StorageType.WORKSPACE),
) -> WorkspaceTree:
    """ Load a Tree from Workspace XML Storage File.

**Parameters:**
 - file_path (Path): The Path to the File containing Workspace XML.

**Returns:**
 WorkspaceTree - The list of Changelist data stored in Workspace Storage.
    """
    if file_path is None:
        exit("Only use Workspace Tree if a Workspace XML file has been given to you.")
    return load_xml(
        file_validation.validate_file_input_text(file_path)
    )


def write_file(
    tree: WorkspaceTree,
    file_path: Path = storage_type.get_default_path(StorageType.WORKSPACE),
) -> bool:
    """ Write a Changelist Data Storage object to an XML File.

**Parameters:**
 - tree (ChangelistDataStorage): The Tree object containing the Changelists.
 - file_path (Path): The File Path to write the XML data to.

**Returns:**
 bool - True after the operation succeeds.
    """
    if tree is None or not isinstance(tree, WorkspaceTree):
        raise TypeError
    return tree.write_tree(file_path)



""" XML Tree File Writing Abstract Class.
"""
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Iterable, Generator
from xml.etree.ElementTree import ElementTree, tostring

from changelist_data.changelist import Changelist


class BaseXMLTree(metaclass=ABCMeta):
    """ A Base Abstract Class providing writing capabilities for an XML Tree class.
    """

    @abstractmethod
    def get_root(self) -> ElementTree:
        raise NotImplementedError

    @abstractmethod
    def get_changelists(self) -> list[Changelist]:
        raise NotImplementedError

    @abstractmethod
    def generate_changelists(self) -> Generator[Changelist, None, None]:
        raise NotImplementedError

    @abstractmethod
    def update_changelists(
        self,
        changelists: list[Changelist] | Iterable[Changelist],
    ): raise NotImplementedError

    def write_tree(
        self, path: Path,
    ) -> bool:
        """ Write the Tree as XML to the given Path.
    - Ensures that all parent directories exist, and creates the file if necessary.

    **Parameters:**
     - path (Path): The Path to the File.

    **Returns:**
     bool - True if data was written to the file.
        """
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch()
        return path.write_bytes(
            tostring(
                element=self.get_root().getroot(),
                encoding='utf-8',
                method='xml',
                xml_declaration=True,
            )
        ) > 0


""" String Validation Methods.
"""
from changelist_data.validation.collections import execute_boolean_operation_concurrently


_ILLEGAL_FILESYSTEM_CHARS: tuple[str,...] = ('&', ';', '\0', '<', '>', '*', '?', '|', '\n', '\r', ',', '^', '#', '@', '$',)


def validate_string_argument(
    argument: str | None,
    check_illegal_fs_chars: bool = True,
) -> bool:
    """ Determine whether an argument is a non-empty string.
 - Does not count whitespace.
 - Uses the strip method to remove empty space.

**Parameters:**
 - argument (str): The given argument to validate.
 - check_illegal_fs_chars (bool): Whether to apply the Illegal Character filter. Default is True.

**Returns:**
 bool - True if the argument qualifies as valid.
    """
    if argument is None or not isinstance(argument, str):
        return False
    elif len(simplified_arg := argument.strip()) < 1:
        return False
    elif check_illegal_fs_chars:
        return not has_illegal_filesystem_chars(simplified_arg)
    return True


def has_illegal_filesystem_chars(
    filesystem_argument: str,
) -> bool:
    """ Determines whether this argument contains illegal filesystem characters.

**Parameters:**
 - filesystem_argument (str): The argument to be validated.

**Returns:**
 bool - True if an illegal character was found in the sequence. The argument must be filtered or rejected.
    """
    for char in _ILLEGAL_FILESYSTEM_CHARS:
        if char in filesystem_argument:
            return True
    return False


def has_illegal_filesystem_chars_concurrent(
    filesystem_argument: str,
) -> bool:
    """ Determines whether this argument contains illegal filesystem characters.

**Parameters:**
 - filesystem_argument (str): The argument to be validated.

**Returns:**
 bool - True if an illegal character was found in the sequence. The argument must be filtered or rejected.
    """
    def check_chars(characters):
        for c in characters:
            if c in filesystem_argument:
                return True
        return False
    return execute_boolean_operation_concurrently(
        _ILLEGAL_FILESYSTEM_CHARS, check_chars, 4
    )



""" Loads the Workspace file into a Tree with read/write capabilities.
"""
from typing import Iterable, Generator
from xml.etree.ElementTree import Element, ElementTree, indent

from changelist_data.changelist import Changelist
from changelist_data.xml.base_xml_tree import BaseXMLTree
from changelist_data.xml.workspace import workspace_writer, workspace_reader


class WorkspaceTree(BaseXMLTree):
    """ Manages the Workspace XML Element Trees.

**Properties:**
 - xml_root (Element): The XML root element.
 - changelist_manager (Element): The Changelist Manager Component Element.
    """

    def __init__(
        self,
        xml_root: Element,
    ):
        self._xml_root = xml_root
        self.changelist_manager = workspace_reader.find_changelist_manager(xml_root)

    def get_changelists(self) -> list[Changelist]:
        """ Obtain the list of CL Elements.

    **Returns:**
     list[Changelist] - A List containing the Lists.
        """
        return list(self.generate_changelists())

    def generate_changelists(self) -> Generator[Changelist, None, None]:
        """ Generate Changelists from the Tree.

    **Yields:**
     Changelist - The data objects extracted from the Storage XML Tree.
        """
        if self.changelist_manager is None:
            exit('XML File does not have a Changelists Element.')
        yield from workspace_reader.generate_changelists(self.changelist_manager)

    def get_root(self) -> ElementTree:
        """ Obtain the XML ElementTree Root.

    **Returns:**
     ElementTree - The XML Tree Root Element.
        """
        return ElementTree(self._xml_root)

    def update_changelists(
        self,
        changelists: Iterable[Changelist],
    ):
        """ Update the XML Tree's Changelist Manager Lists.

    **Parameters:**
     - changelists (Iterable[Changelist]): The List or Iterable of Changelists.
        """
        if (clm := self.changelist_manager) is None:
            exit('XML File does not have a Changelist Manager.')
        # First obtain all Option Elements
        options = list(clm.findall('option'))
        # Clear the Changelist Manager Tag
        clm.clear() # Need to Add Name Attribute after Clear operation
        clm.attrib['name'] = "ChangeListManager"
        # Add All Sub Elements
        clm.extend(workspace_writer.write_list_element(x) for x in changelists)
        clm.extend(options)
        indent(clm, level=1)


""" Storage Access Methods.
"""
from pathlib import Path
from typing import Generator

from changelist_data.changelist import Changelist
from changelist_data.storage import file_validation, changelists_storage, workspace_storage
from changelist_data.storage.changelist_data_storage import ChangelistDataStorage
from changelist_data.storage.storage_type import StorageType, get_default_path
from changelist_data.xml.changelists import new_tree


def read_storage(
    option: StorageType | None = None,
    file_path: Path | None = None,
) -> list[Changelist]:
    """ Read Changelist Data from Storage into a List of Changelist data.
 - None values indicate that the storage file should be searched for using default values.

**Parameters:**
 - option (StorageType?): The Storage Type describing the XML format. Search all StorageTypes in order by default. Default: None.
 - file_path (Path?): The file path to read from. Use None to look in default storage locations. Default: None.

**Returns:**
 list[Changelist] - The List of Changelist data.
    """
    if file_path is not None:
        if option is None:
            exit("When File Path is Provided, The Storage Type must be Provided as well.")
        return list(_generate_option(option, file_path))
    # Find and Read from Default File Paths
    if option is None:
        if (storage_result := read_any_storage_option()) is not None:
            return storage_result
    elif (storage_path := file_validation.check_if_default_file_exists(option)) is not None:
        return list(_generate_option(option, storage_path))
    return []


def read_any_storage_option() -> list[Changelist] | None:
    """ Check the default locations for all storage options.
 - First checks Changelists, then Workspace locations.

**Returns:**
 list[Changelist]? - The list of Changelist data objects read from the file, or None.
    """
    if len(result := list(_generate_changelists_from_any_storage_option())) > 0:
        return result
    return None


def _generate_changelists_from_any_storage_option() -> Generator[Changelist, None, None]:
    for opts in StorageType:
        if (storage_path := file_validation.check_if_default_file_exists(StorageType(opts))) is not None:
            yield from _generate_option(StorageType(opts), storage_path)
            break
    return None


def generate_changelists_from_storage(
    option: StorageType | None = None,
    file_path: Path | None = None,
) -> Generator[Changelist, None, None]:
    """ Read Changelist Data from Storage into a List of Changelist data.
 - None values indicate that the storage file should be searched for using default values.

**Parameters:**
 - option (StorageType?): The Storage Type describing the XML format. Search all StorageTypes in order by default. Default: None.
 - file_path (Path?): The file path to read from. Use None to look in default storage locations. Default: None.

**Returns:**
 list[Changelist] - The List of Changelist data.
    """
    if file_path is not None:
        if option is None:
            exit("When File Path is Provided, The Storage Type must be Provided as well.")
        yield from _generate_option(option, file_path)
        return None
    # Find and Read from Default File Paths
    if option is None:
        if (storage_result := _generate_changelists_from_any_storage_option()) is not None:
            yield from storage_result
            return None
    elif (storage_path := file_validation.check_if_default_file_exists(option)) is not None:
        yield from _generate_option(option, storage_path)
    return None


def load_storage(
    option: StorageType | None = None,
    file_path: Path | None = None,
) -> ChangelistDataStorage:
    """ Load Changelist Data into a managed Storage object.
 - None values indicate that the storage file should be searched for using default values.

**Parameters:**
 - option (StorageType?): The Storage Type describing the XML format. Use None to search all StorageTypes in order. Default: None.
 - file_path (Path?): The file path to read from. Use None to look in default storage locations. Default: None.

**Returns:**
 ChangelistDataStorage - A managed Storage object containing Changelist data.
    """
    if file_path is not None:
        if option is None:
            exit("When File Path is Provided, The Storage Type must be Provided as well.")
        return _load_option(option, file_path)
    # Find and Read from Default File Paths
    if option is None:
        if (storage_result := load_any_storage_option()) is not None:
            return storage_result
    elif (storage_path := file_validation.check_if_default_file_exists(option)) is not None:
        return _load_option(option, storage_path)
    # Create an empty Changelists Storage Tree with a default path
    return ChangelistDataStorage(
        new_tree(),
        StorageType.CHANGELISTS,
        get_default_path(StorageType.CHANGELISTS)
    )


def load_any_storage_option() -> ChangelistDataStorage | None:
    """ Check the default locations for all storage options.
 - First checks Changelists, then Workspace locations.
 - Returns None if no file is found.

**Returns:**
 ChangelistDataStorage? - The data object providing a read and write interface for the storage file.
    """
    for opts in StorageType:
        if (storage_path := file_validation.check_if_default_file_exists(StorageType(opts))) is not None:
            return _load_option(StorageType(opts), storage_path)
    return None


def _generate_option(
    option: StorageType,
    path: Path,
) -> Generator[Changelist, None, None]:
    if option == StorageType.CHANGELISTS:
        yield from changelists_storage.generate_changelists_from_file(path)
    elif option == StorageType.WORKSPACE:
        yield from workspace_storage.generate_changelists_from_file(path)
    return None


def _load_option(
    option: StorageType,
    path: Path,
) -> ChangelistDataStorage:
    return ChangelistDataStorage(
        changelists_storage.load_file(path) if option == StorageType.CHANGELISTS else workspace_storage.load_file(path),
        option,
        path
    )



""" File Validation Methods.
"""
from pathlib import Path

from changelist_data.storage.storage_type import StorageType, get_default_path


def file_exists(
    path: Path,
) -> bool:
    """ Determines if Path exists and is_file.

**Parameters:**
 - path (Path): The path to check.

**Returns:**
 bool - Whether the file exists.
    """
    return path.exists() and path.is_file()


def check_if_default_file_exists(
    option: StorageType,
) -> Path | None:
    """ Return the Path to the default file for this storage option iff it exists.

**Parameters:**
 - option (StorageType): The StorageType option to check.

**Returns:**
 Path? - The Path to the Default Storage File, or None if the path is not a file that exists.
    """
    if file_exists(file_path := get_default_path(option)):
        return file_path
    return None


def validate_file_input_text(
    file_path: Path,
) -> str:
    """ Ensure that the File Exists, and is within reasonable size parameters.
 - Read the File and return its string contents.

**Parameters:**
 - file_path (Path): The Path to the Input File.

**Returns:**
 str - The Text Contents of the Input File.

**Raises:**
 SystemExit - When any of the validation conditions fails, or the file operation fails.
    """
    try:
        if not file_path.exists():
            exit("File did not exist.")
        if not file_path.is_file():
            exit("Given Path was not a file.")
        if (file_size := file_path.stat().st_size) > 32 * 1024**2:
            exit("Input File was larger than 32 MB. Refusing to read it.")
        return file_path.read_text()
    except UnicodeError:
        exit("File is not valid text. Unicode error.")
    except OSError:
        exit("OSError occurred while reading Input File.")
    except Exception as e:
        exit(f"Unexpected Exception while reading InputFile(name={file_path.name}, fileSize={file_size / 1024} kb) Exception=({e}).")


""" Reads the Changelists File, translates into Changelists Data types.
"""
from typing import Generator, Iterable
from xml.etree.ElementTree import Element, ParseError, fromstring
from changelist_data.changelist import Changelist
from changelist_data.file_change import FileChange
from changelist_data.xml import xml_reader


def parse_xml(changelists_xml: str) -> Element:
    """ Parse an XML File. This should be a Changelists XML file.

    Parameters:
    - changelists_xml (str): The Changelists data in xml format.

    Returns:
    Element - the XML Root Element

    Raises:
    SystemExit - if the xml could not be parsed.
    """
    try:
        return fromstring(changelists_xml)
    except ParseError:
        exit("Unable to Parse Changelists XML File.")


def find_changelists_root(xml_root: Element) -> Element | None:
    """ Extract the ChangeLists Root XML Element.

    Parameters:
    - xml_root (Element): The parsed xml root.
    
    Returns:
    Element - The XML Changelists element, or None.
    """
    for elem in xml_reader.filter_by_tag(xml_root, 'changelists'):
        return elem
    return None


def extract_list_elements(changelists_element: Element) -> list[Changelist]:
    """ Given the Changelist Manager Element, obtain the list of List Elements.

    Parameters:
    - changelist_manager (Element): The ChangeList Manager XML Element.

    Returns:
    list[Element] - A List containing the Lists.
    """
    return list(generate_changelists(changelists_element))


def generate_changelists(
    cl_container: Element
) -> Generator[Changelist, None, None]:
    """ Generate Changelists from the XML Tree Element.
- The given Element's sub-elements are filtered by tag for 'list'.
- These 'list' elements are translated into Changelist data objects.

**Parameters:**
 - cl_container (Element): The XML Changelist Container Element.

**Yields:**
 Changelist - The CL Data objects extracted from the XML tree.
    """
    for cl_element in xml_reader.filter_by_tag(cl_container, 'list'):
        yield Changelist(
            id=xml_reader.get_attr(cl_element, 'id'),
            name=xml_reader.get_attr(cl_element, 'name'),
            changes=_extract_change_data(cl_element),
            comment=xml_reader.get_attr_or(cl_element, 'comment', ''),
            is_default=xml_reader.read_bool_from(cl_element, 'default'),
        )


def _extract_change_data(
    list_element: Element,
) -> list[FileChange]:
    """ Given a ChangeList XML Element, obtain the List of Changes.

    Parameters:
    - list_element (Element): The Element representing a Changelist.

    Returns:
    list[FileChange] - The list of structured FileChange.
    """
    return list(_generate_change_data(
         filter(lambda x: x.tag == 'change', list_element)
    ))


def _generate_change_data(
    elements: Iterable[Element],
) -> Generator[FileChange, None, None]:
    for change in elements:
        yield FileChange(
            before_path=xml_reader.get_attr(change, 'beforePath'),
            before_dir=xml_reader.convert_bool(xml_reader.get_attr(change, 'beforeDir')),
            after_path=xml_reader.get_attr(change, 'afterPath'),
            after_dir=xml_reader.convert_bool(xml_reader.get_attr(change, 'afterDir')),
        )


""" The Options for Changelists data Storage.
"""
from enum import Enum
from pathlib import Path


class StorageType(Enum):
    """ The Type of Storage File to be used by Changelists.
    """
    CHANGELISTS = "changelists"
    WORKSPACE = "workspace"


CHANGELISTS_FILE_PATH_STR = '.changelists/data.xml'
WORKSPACE_FILE_PATH_STR = '.idea/workspace.xml'


def get_default_file(storage_type: StorageType) -> str:
    """ Get the Default File Location for this Storage Type, as a String.

**Parameters:**
 - storage_type (StorageType): The enum describing which Storage to use.

**Returns:**
 str - The file location as a string.
    """
    if storage_type == StorageType.CHANGELISTS:
        return CHANGELISTS_FILE_PATH_STR
    if storage_type == StorageType.WORKSPACE:
        return WORKSPACE_FILE_PATH_STR
    # Add New Enums Here:
    raise ValueError(f"Invalid Argument: {storage_type}")


def get_default_path(storage_type: StorageType) -> Path:
    """ Get the Default File Path for this Storage Type.

**Parameters:**
 - storage_type (StorageType): The enum describing which Storage to use.

**Returns:**
 Path - The file Path object.
    """
    return Path(get_default_file(storage_type))



""" Changelist Data Package.
"""
from pathlib import Path
from typing import Generator

from changelist_data.changelist import Changelist
from changelist_data.storage import load_storage, load_any_storage_option, read_storage, read_any_storage_option, \
    generate_changelists_from_storage, _generate_changelists_from_any_storage_option
from changelist_data.storage.changelist_data_storage import ChangelistDataStorage
from changelist_data.storage.storage_type import StorageType, get_default_path
from changelist_data.validation.arguments import validate_string_argument
from changelist_data.xml.changelists import new_tree


def read_storage_from_file_arguments(
    changelists_file: str | None = None,
    workspace_file: str | None = None,
) -> list[Changelist] | None:
    """ Process the Given File Arguments, and read from Storage.
 - Storage is managed by changelist_data package.

**Parameters:**
 - changelists_file (str?): A string path to the Changelists file, if specified.
 - workspace_file (str?): A string path to the Workspace file, if specified.

**Returns:**
 list[Changelist] - The Changelist data from the storage file.
    """
    if validate_string_argument(changelists_file):
        return read_storage(StorageType.CHANGELISTS, Path(changelists_file))
    if validate_string_argument(workspace_file):
        return read_storage(StorageType.WORKSPACE, Path(workspace_file))
    return read_any_storage_option()


def generate_changelists_from_storage_file_arguments(
    changelists_file: str | None = None,
    workspace_file: str | None = None,
) -> Generator[Changelist, None, None]:
    """ Process the Given File Arguments, and generate Changelist data objects from Storage.
 - Storage is managed by changelist_data.storage package.

**Parameters:**
 - changelists_file (str?): A string path to the Changelists file, if specified.
 - workspace_file (str?): A string path to the Workspace file, if specified.

**Yields:**
 Generator[Changelist] - The Changelist data from the storage file.
    """
    if validate_string_argument(changelists_file):
        return generate_changelists_from_storage(StorageType.CHANGELISTS, Path(changelists_file))
    if validate_string_argument(workspace_file):
        return generate_changelists_from_storage(StorageType.WORKSPACE, Path(workspace_file))
    return _generate_changelists_from_any_storage_option()


def load_storage_from_file_arguments(
    changelists_file: str | None = None,
    workspace_file: str | None = None,
) -> ChangelistDataStorage:
    """ Use the given optional parameters and search for the storage file.
 - Changelists_file argument overrides Workspace_file.
 - Defaults to searching Changelists then Workspace file default locations.

**Parameters:**
 - changelists_file (str?): The string path to the Changelists XML Data file.
 - workspace_file (str?): The string path to the Workspace XML file.

**Returns:**
 ChangelistDataStorage - The data class container for Changelist information.
    """
    if validate_string_argument(changelists_file):
        return load_storage(StorageType.CHANGELISTS, Path(changelists_file))
    if validate_string_argument(workspace_file):
        return load_storage(StorageType.WORKSPACE, Path(workspace_file))
    if (existing_storage := load_any_storage_option()) is not None:
        return existing_storage
    # Create a new Changelists Storage file
    return ChangelistDataStorage(
        new_tree(),
        StorageType.CHANGELISTS,
        get_default_path(StorageType.CHANGELISTS)
    )


""" Changelist Data Storage XML Workspace
"""
from typing import Generator

from changelist_data.changelist import Changelist
from changelist_data.xml.workspace import workspace_reader
from changelist_data.xml.workspace.workspace_tree import WorkspaceTree


def read_xml(
    workspace_xml: str
) -> list[Changelist]:
    """ Parse the Workspace XML file and obtain all ChangeList Data in a list.

**Parameters:**
 - workspace_xml (str): The contents of the Workspace file, in xml format.
    
**Returns:**
 list[Changelist] - The list of Changelists in the workspace file.
    """
    return list(generate_changelists_from_xml(workspace_xml))


def generate_changelists_from_xml(
    workspace_xml: str
) -> Generator[Changelist, None, None]:
    """ Generates Changelist Data from the Given XML String.
 - Parses the XML string using ElementTree.
 - Creates Changelist data objects one at a time from the ElementTree.

**Parameters:**
 - workspace_xml (str): The Workspace xml file contents as a string.

**Yields:**
 Changelist - The Changelist data objects extracted from the XML ElementTree.
    """
    if (cl_manager := workspace_reader.find_changelist_manager(
        workspace_reader.parse_xml(workspace_xml)
    )) is None:
        exit("Changelist Manager was not found in the workspace file.")
    yield from workspace_reader.generate_changelists(cl_manager)


def load_xml(
    workspace_xml: str
) -> WorkspaceTree:
    """ Parse the Workspace XML file into an XML Tree, and Wrap it.

**Parameters:**
 - workspace_xml (str): The contents of the Workspace file, in xml format.

**Returns:**
 WorkspaceTree - An XML Tree changelists interface.
    """
    return WorkspaceTree(
        workspace_reader.parse_xml(workspace_xml)
    )



""" An Abstract Class defining the interface for translation between XML Trees and Changelists.
"""
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Generator

from changelist_data.changelist import Changelist
from changelist_data.storage.storage_type import StorageType
from changelist_data.xml.base_xml_tree import BaseXMLTree


@dataclass(frozen=True)
class ChangelistDataStorage:
    """ Controller Interface for Data Storage.

**Fields:**
 - base_xml_tree (BaseXMLTree): A Wrapper around XML ElementTree, compatible with both xml format Storage Types.
 - storage_type (StorageType): An enum selecting the specific xml format used.
 - update_path (Path): The file Path where Data is read from and written to.
    """
    base_xml_tree: BaseXMLTree
    storage_type: StorageType
    update_path: Path

    def get_changelists(self) -> list[Changelist]:
        """ Obtain a list of the Changelists from the Storage xml.

**Returns:**
 list[Changelist] - A list of Changelist data objects from the Storage xml file.
        """
        return self.base_xml_tree.get_changelists()

    def generate_changelists(self) -> Generator[Changelist, None, None]:
        """ Generate Changelist data objects from the Storage xml tree.

**Yields:**
 Changelist - The Changelist data objects extracted from the XML tree.
        """
        return self.base_xml_tree.generate_changelists()

    def update_changelists(
        self, changelists: Iterable[Changelist],
    ):
        """ Overwrite the collection of Changelist data in Memory.

**Parameters:**
 - changelists (Iterable[Changelist]): The Changelists to insert into the Storage XML Tree.
        """
        self.base_xml_tree.update_changelists(changelists)

    def write_to_storage(self) -> bool:
        """ Create or overwrite storage file.
 - Ensures parent directories exist.

**Returns:**
 bool - True if data was written.
        """
        return self.base_xml_tree.write_tree(self.update_path)



"""XML Reader Methods.
"""
from typing import Iterator
from xml.etree.ElementTree import Element


def filter_by_tag(element: Element, tag: str) -> Iterator[Element]:
    """ Obtain all Elements that match a given Tag.
    
**Parameters:**
 - element (Element): The element to search within.
 - tag (str): The name of the Tag to filter for.

**Returns:**
 Iterator[Element] - An iterable object providing XML Elements.
    """
    return filter(lambda x: x.tag == tag, element.iter())


def get_attr(element: Element, attr: str) -> str | None:
    """ Obtain the Attribute from an element safely, returning None.

**Parameters:**
 - element (Element): The element to obtain the attribute from.
 - attr (str): The name of the attribute.

**Returns:**
 str? - The value of the attribute, or None.
    """
    return element.attrib[attr] if attr in element.attrib else None


def get_attr_or(element: Element, attr: str, default: str) -> str:
    """ Obtain the Attribute from an element safely, returning None if not found.

**Parameters:**
- element (Element): The element to obtain the attribute from.
- attr (str): The name of the attribute.

**Returns:**
 str - The value of the attribute in the element, or the default.
    """
    return element.attrib[attr] if attr in element.attrib else default


def read_bool_from(element: Element, attr: str) -> bool:
    """ Obtain the Attribute from an element safely, returning False if not found.

**Parameters:**
 - element (Element): The element to obtain the attribute from.
 - attr (str): The name of the attribute.

**Returns:**
 bool - The attribute converted to a boolean.
    """
    return get_attr_or(element, attr, '').lower() == 'true'


def convert_bool(attr: str | None) -> bool | None:
    """ Convert a String attribute to boolean.
    """
    return None if attr is None else attr == 'true'



""" Validation Module Methods for Collections.
"""
from typing import Generator, Callable


def is_collection(input_argument) -> bool:
    """ Determines whether the input argument is a collection.
 - Valid collections are either a list or a tuple.

**Parameters:**
 - input_argument (list | tuple): The argument to type-check.

**Returns:**
 bool - True when the argument is a list or a tuple.
    """
    return isinstance(input_argument, list) or isinstance(input_argument, tuple)


def divide_collection(
    input_collection: list | tuple,
    num_groups: int = 2,
) -> Generator[list, None, None]:
    """ Divides a collection into smaller lists, yielding the results one by one.
 - Empty Input returns nothing.

**Parameters:**
 - input_list (list | tuple[...]): The collection that will be divided.
 - num_groups (int): The number of groups to divide the collection into. Must be non-zero positive integer. Default: 2.

**Yields:**
 list - Only yields lists of elements, even if a tuple was provided.
    """
    # Validate Arguments
    if not is_collection(input_collection):
        raise ValueError("Input must be a collection (list or tuple), was " + str(type(input_collection)))
    if not isinstance(num_groups, int) or num_groups <= 0:
        raise ValueError("Number of groups must be a positive integer")
    # Check the Numbers
    if num_groups > (collection_size := len(input_collection)):
        if 0 == (num_groups := collection_size):
            return
    group_size = collection_size // num_groups # Calculate the size of each group
    for i in range(num_groups):
        start_index = i * group_size
        end_index = (i + 1) * group_size if i < num_groups - 1 else None
        yield input_collection[start_index:end_index]


def execute_boolean_operation_concurrently(
    input_collection: list | tuple,
    collection_operation: Callable[[list | tuple], bool],
    num_concurrent_groups: int = 4,
) -> bool:
    """ Execute a boolean operation on a collection of elements, concurrently.
 - Returns True as soon as possible.

**Parameters:**
 - input_collection (list | tuple): The collection to be split and operated on concurrently.
 - collection_operation (Callable): A callable that executes on a list, and returns the first True result.
 - num_concurrent_groups (int): The number of groups to split the collection into. Default: 4.

**Returns:**
 bool - True if any element in the collection causes the operation to return true.
    """
    from concurrent.futures import ThreadPoolExecutor, as_completed
    #
    with ThreadPoolExecutor() as executor:
        futures = [ # Submit tasks for groups of illegal characters
            executor.submit(collection_operation, char_groups)
            for char_groups in divide_collection(input_collection, num_concurrent_groups)
        ]
        for future in as_completed(futures): # As each task completes check for true result
            if future.result():
                return True
    return False



""" The Data Class for a ChangeList.
"""
from dataclasses import dataclass, field
from typing import Iterable

from changelist_data.file_change import FileChange


@dataclass(frozen=True)
class Changelist:
    """ The Data class representing a ChangeList.
    
**Properties:**
 - id (str): The unique id of the changelist.
 - name (str): The name of the changelist.
 - changes (list[FileChange]): The list of FileChange data objects. Default: Empty List.
 - comment (str): The comment associated with the changelist. Default: Empty String.
 - is_default (bool): Whether this is the active changelist. Default: False.
    """
    id: str
    name: str
    changes: list[FileChange] = field(default_factory=lambda: [])
    comment: str = ""
    is_default: bool = False


def get_default_cl(
    changelists: list[Changelist] | Iterable[Changelist],
) -> Changelist | None:
    """ Find the Default Changelist, or the first Changelist.
- Returns None if collection argument is empty.
- Handles list arguments slightly differently from Iterables.

**Parameters:**
 - changelists (list | Iterable[Changelist]): The collection of Changelist data objects.

**Returns:**
 Changelist? - The Default, or first CL, or None if the collection was empty.
    """
    if isinstance(changelists, list):
        try:
            return filter(lambda x: x.is_default, changelists).__next__()
        except StopIteration:
            pass
        try:
            return changelists[0]
        except IndexError:
            return None
    elif isinstance(changelists, Iterable):
        try:
            iterator = changelists.__iter__()
            if (first := iterator.__next__()).is_default:
                return first
            try:
                return filter(lambda x: x.is_default, iterator).__next__()
            except StopIteration:
                return first
        except StopIteration:
            return None
    else:
        raise TypeError(type(changelists))


def compute_key(cl_name: str) -> str:
    """ Compute a Key to use for a given Changelist Name.
 - computation is a sequence of reduction operations

**Parameters:**
 - cl_name (str): The Changelist Name to use in key computation.

**Returns:**
 str - A Key that can be used for changelist lookups, or an empty str.
    """
    if len(cl_name) == 0:
        return ''
    translator = str.maketrans('', '', ' :/\\')
    words = cl_name.translate(translator).split()
    return ''.join(w.lower() for w in words)